﻿// Copyright (c) 2020 Tudor Stanciu

using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http.Headers;
using System.Security.Claims;
using System.Text.Encodings.Web;
using System.Threading.Tasks;
using Tuitio.Application.Services.Abstractions;
using Tuitio.Domain.Models;

namespace Tuitio.Authentication
{
    public class AuthenticationHandler : AuthenticationHandler<AuthenticationSchemeOptions>
    {
        private readonly IUserService _userService;
        private readonly ILogger<AuthenticationHandler> _logger;

        public AuthenticationHandler(IOptionsMonitor<AuthenticationSchemeOptions> options, ILoggerFactory loggerFactory, UrlEncoder encoder, ISystemClock clock, IUserService userService, ILogger<AuthenticationHandler> logger) : base(options, loggerFactory, encoder, clock)
        {
            _userService=userService;
            _logger=logger;
        }

        protected override async Task<AuthenticateResult> HandleAuthenticateAsync()
        {
            var endpoint = Context.GetEndpoint();
            if (endpoint?.Metadata?.GetMetadata<IAllowAnonymous>() != null)
            {
                return AuthenticateResult.NoResult();
            }

            var token = GetAuthorizationToken();
            if (token == null)
                return AuthenticateResult.Fail("AUTHORIZATION_HEADER_IS_MISSING");

            var result = await Task.Run(() => _userService.Authorize(token));
            if (result == null)
                return AuthenticateResult.Fail("UNAUTHORIZED");

            var ticket = GetAuthenticationTicket(result);
            return AuthenticateResult.Success(ticket);
        }

        private string GetAuthorizationToken()
        {
            if (Request.Headers.ContainsKey("Authorization"))
            {
                var authorizationHeader = AuthenticationHeaderValue.Parse(Request.Headers["Authorization"]);
                var token = authorizationHeader.Parameter;
                return token;
            }

            return null;
        }

        private AuthenticationTicket GetAuthenticationTicket(TokenExtension result)
        {
            var claimCollection = new Dictionary<string, string>()
            {
                { ClaimTypes.NameIdentifier, result.Token.UserId.ToString() },
                { ClaimTypes.Name, result.Token.UserName },
            };

            if (result.Token.FirstName != null)
                claimCollection.Add(ClaimTypes.GivenName, result.Token.FirstName);

            if (result.Token.LastName != null)
                claimCollection.Add(ClaimTypes.Surname, result.Token.FirstName);

            if (result.Token.Email != null)
                claimCollection.Add(ClaimTypes.Email, result.Token.Email);

            if (result.Token.Claims != null && result.Token.Claims.Any())
            {
                foreach (var claim in result.Token.Claims)
                {
                    if (claimCollection.ContainsKey(claim.Key))
                    {
                        _logger.LogWarning($"There is already a claim with key {claim.Key} in the collection. The combination {claim.Key}:{claim.Value} will be ignored.");
                        continue;
                    }

                    claimCollection.Add(claim.Key, claim.Value);
                }
            }

            var claims = claimCollection.Select(z => new Claim(z.Key, z.Value)).ToArray();
            var identity = new ClaimsIdentity(claims, Scheme.Name);
            var principal = new ClaimsPrincipal(identity);
            var ticket = new AuthenticationTicket(principal, Scheme.Name);

            return ticket;
        }
    }
}
